@@ -1,11 +1,12 @@
 /*****************************************************************************
- * Secure Locate v1.6                                                        *
- * May 28, 1999                                                              *
+ * Secure Locate v2.0                                                        *
+ * July 1, 1999                                                              *
  * Programmed by: Kevin Lindsay                                              *
  * Copyright (c) 1998 Kevin Lindsay & Netnation Communications Inc.          *
  *                                                                           *
- * NetNation Communications:  http://www.netnation.com/                      *
+ * NetNation Communications   http://www.netnation.com/                      *
  * Secure Locate: ftp://ftp.mkintraweb.com/pub/linux/slocate/                *
+ *                http://pep.netnation.com/slocate/                          *
  *                                                                           *
  * Patches:  Sean Mcnulty <lazy@ccf.rutgers.edu>                             *
  *                * Fixed a bug which caused the decode function to fail.    *
@@ -32,6 +33,10 @@
  *                                                                           *
  * Even though Secure Locate was written from scratch, some recognition goes *
  * to James A. Woods <jwoods@adobe.com> for writing the original GNU locate. *
+ *                                                                           *
+ * This product includes software developed by the University of             *
+ * California, Berkeley and its contributors.                                *
+ *                                                                           *
  *****************************************************************************/
 
 /*****************************************************************************
@@ -71,17 +76,27 @@
 #include <ctype.h>
 #include <fnmatch.h>
 #include <regex.h>
+#include "sl_fts.h"
 
 #include "link.h"
+#include "misc.h"
+
+#ifdef __FreeBSD__
+# include <sys/param.h>
+# include <sys/ucred.h>
+# include <sys/mount.h>
+#endif
 
 /* GLOBALS */
 
-#define VERSION "Secure Locate v1.6 - Released May 28, 1999\n"
+#define VERSION "Secure Locate v2.0 - Released July 1, 1999\n"
 #define GRPFILE "/etc/group"
 #define SLOC_ESC 0x80
 #define SLOC_GRP "slocate"
 #define SLOC_UID 0
 #define MIN_BLK 64
+#define MTAB_FILE "/etc/mtab"
+
 /* Warn if a database is older than this.  8 days allows for a weekly
  *    update that takes up to a day to perform.  */
 #define WARN_SECONDS (60 * 60 * 24 * 8)
@@ -94,19 +109,21 @@ char **SLOCATE_PATH = NULL;
 /* More fitting paths for FreeBSD -matt */
 #if defined(__FreeBSD__)
 char *SLOCATEDB = "/var/db/slocate/slocate.db";
-char *TMPSLOCATEDB = "/var/db/slocate/slocate.tmp";
+char *TMPSLOCATEDB = "/var/db/slocate/slocate.db.tmp";
 char *SLOCATEDB_DIR = "/var/db/slocate/";
 #else
 /* These paths are good for Linux machines */
 char *SLOCATEDB = "/var/lib/slocate/slocate.db";
-char *TMPSLOCATEDB = "/var/lib/slocate/slocate.tmp";
+char *TMPSLOCATEDB = "/var/lib/slocate/slocate.db.tmp";
 char *SLOCATEDB_DIR = "/var/lib/slocate/";
+#define UPDATEDB_CONF "/etc/updatedb.conf"
 #endif
-char *EXCLUDE_DIR;
+char *EXCLUDE_DIR=NULL;
 int EXCLUDE=0;
 int VERBOSE=0;
 int QUIET=0;
 int REGEXP=0;
+int NEWOUTPUT=0;
 char *regexp;
 char *progname;
 char *tmp_path=NULL;
@@ -128,23 +145,45 @@ int decode_db(char *database, char *str);
 void
 usage() 
 {
+   int i;
    printf("%s\n"
           "Copyright (c) 1999 Kevin Lindsay & Netnation Communications Inc.\n\n"
           "search usage:   %s [-q] [-d <path>] [--database=<path>] <search string>...\n"
           "                %s [-r <regexp>] [--regexp=<regexp>]\n"
           "database usage: %s [-qv] [-o <file>] [--output=<file>]\n"
-          "                %s [-e <dir1,dir2,...>] [-l <level>] <[-U <path>] [-u]>\n"
+          "                %s [-e <dir1,dir2,...>] [-f <fs_type1,...> ] [-l <level>]\n"
+          ,VERSION,progname,progname,progname,progname);
+   
+   for (i = 0; i < strlen(progname)-1; i+=1)
+     printf(" ");
+   
+   printf(
+#ifndef __FreeBSD__
+          "                  [-c] <[-U <path>] [-u]>\n"
+#else
+          "                  <[-U <path>] [-u]>\n"          
+#endif
           "general usage:  %s [-Vh] [--version] [--help]\n\n"
           "   Options:\n"
           "   -u                 - Create slocate database starting at path /.\n"
-          "   -U <dir>           - Create slocate database starting at path <dir>.\n"          
-          "   -e <dir1,dir2,...> - Exclude directories from the slocate database.\n"
+          "   -U <dir>           - Create slocate database starting at path <dir>.\n",progname);
+#ifndef __FreeBSD__
+   printf("   -c                 - Parse original GNU Locate's '/etc/updatedb.conf'\n"
+          "                        when using the -u or -U options.  If 'updatedb' is\n"
+          "                        symbolically linked to the '%s' binary, the\n"
+          "                        original configuration file will automatically be\n"
+          "                        used.\n",progname);
+#endif
+   printf("   -e <dir1,dir2,...> - Exclude directories from the slocate database when\n"
+          "                        using the -u or -U options.\n"
+          "   -f <fs_type1,...>  - Exclude file system types from the slocate database\n"
+          "                        when using the -u or -U options. (ie. NFS, etc).\n"                                  
           "   -l <level>         - Security level. \n"
           "                           0 turns security checks off. This will make\n"
           "                             searchs faster.\n"
           "                           1 turns security checks on. This is the default.\n"
           "   -q                 - Quiet mode.  Error messages are suppressed.\n"
-          "   -n <num>            - Limit the amount of results shown to <num>.\n"
+          "   -n <num>           - Limit the amount of results shown to <num>.\n"
           "   -r <regexp>\n"
           "   --regexp=<regexp>  - Search the database using a basic POSIX regular\n"
           "                        expression.\n"
@@ -161,37 +200,13 @@ usage()
           "\n"
           "Author: Kevin Lindsay\n"
           "Bugs:   klindsay@mkintraweb.com\n"
-          "FTP:    ftp.mkintraweb.com/pub/linux/slocate/\n"
-          "\n",VERSION,progname,progname,progname,progname,progname);
+          "FTP:    ftp://ftp.mkintraweb.com/pub/linux/slocate/\n"
+          "HTTP:   http://pep.netnation.com/slocate/\n"
+          "\n");
    
    exit(0);
 }
 
-/* My fchdir() so that I can do cleaner error checking */
-
-int
-MYfchdir(int fd, char *filepath, char *func)
-{   
-   if (fchdir(fd) == -1) {
-      if (!QUIET)
-          fprintf(stderr,"%s: %s: fchdir: %s: %s\n",progname,func,filepath,strerror(errno));
-      return(0);
-   }
-   return(1);
-}
-
-/* My chdir() so that I can do cleaner error checking */
-int
-MYchdir(char *filepath, char *func)
-{
-   if (chdir(filepath) == -1) {
-      if (!QUIET)      
-          fprintf(stderr,"%s: %s: chdir: %s: %s\n",progname,func,filepath,strerror(errno));
-      return(0);
-   }
-   return(1);
-}
-
 static void
 put_short (int c, FILE *fp)
 {
@@ -213,16 +228,18 @@ get_short (FILE *fp)
 void
 parse_create_path(char *path)
 {
-   struct stat statres;
-
+   int ret_val;
+   
    if (!path || strlen(path) == 0) return;
-
-   stat(path, &statres);
-   if (S_ISDIR(statres.st_mode)) {
-   if (!QUIET)
-        fprintf(stderr,"%s: %s is a directory\n", progname,path);
-   exit(1);
+   
+   ret_val = creat(path, S_IRWXU|S_IRGRP);
+   if(ret_val==-1)
+   {
+      if (!QUIET)
+        printf("%s: could not create database: %s: %s\n", progname,path,strerror(errno));
+      exit(1);
    }
+      
    SLOCATEDB = malloc(strlen(path)+1);
    SLOCATEDB[0] = '\0';
    strcat(SLOCATEDB, path);
@@ -241,22 +258,22 @@ parse_decode_path(char *path)
    char *pathcopy;
    char *part;
    int i;
-
+   
    if (!path || strlen(path) == 0) return;
-
+   
    i = 1;
    part = path;
    while ((part = strchr(part+1, ':'))) i++;
    SLOCATE_PATH = malloc(i * sizeof(char *));
-
+   
    i = 0;
    pathcopy = malloc(strlen(path)+1);
    pathcopy[0] = '\0';
    strcat(pathcopy, path);
    part = strtok(pathcopy, ":");
    while(part) {
-   SLOCATE_PATH[i++] = part;
-   part = strtok(NULL, ":");
+      SLOCATE_PATH[i++] = part;
+      part = strtok(NULL, ":");
    }
    SLOCATE_PATH[i] = NULL;
 }
@@ -290,6 +307,7 @@ parse_dash(char *option)
 
    if (!strcmp(option,"OUTPUT")) {
       parse_create_path(ptr);
+      NEWOUTPUT=1;
    }
 
    if (!strcmp(option,"DATABASE")) {
@@ -348,17 +366,17 @@ get_gid()
 
 /* Parse Exclude Command */
 
-char *
+int
 parse_exclude(char *estr)
 {
    char *excludestr=NULL;
    char *ptr1;
    char *ptr2=NULL;
    int grow=1;
-   int elen;
-   
-   EXCLUDE=1;
+   int elen;   
 
+   EXCLUDE=1;   
+   
    excludestr = (char *)malloc(1);
    excludestr[0] = '\0';
    
@@ -374,7 +392,7 @@ parse_exclude(char *estr)
       }
 
       grow += (ptr2-ptr1)+2;
-      if (*(ptr2-1) == '/') {
+      if (*(ptr2-1) == '/' && (ptr2-1) != estr && *(ptr2-2) != ',' ) {
          grow--;
          ptr2--;
       }
@@ -391,9 +409,353 @@ parse_exclude(char *estr)
           ptr1 = ptr2+1;
    }
    
-   return (excludestr);
+   if (!EXCLUDE_DIR) {
+      EXCLUDE_DIR = malloc(strlen(excludestr)+1);
+      *EXCLUDE_DIR = 0;      
+   } else {
+      EXCLUDE_DIR = realloc(EXCLUDE_DIR,strlen(EXCLUDE_DIR)+strlen(excludestr)+1);
+   }
+   
+   strcat(EXCLUDE_DIR,excludestr);
+   free(excludestr);
+
+#ifdef DEBUG
+   printf("E: %s\n",EXCLUDE_DIR);
+#endif
+
+   return 1;
 }
 
+
+#ifdef __FreeBSD__
+/* Get File System type in the form of a string. "*fstype*" */
+
+char *
+get_fs_type(int fs_type)
+{
+   if (fs_type == MOUNT_UFS)
+     return("*UFS*");
+   else if (fs_type == MOUNT_NFS)
+     return("*NFS*");
+   else if (fs_type == MOUNT_MFS)
+     return("*MFS*");
+   else if (fs_type == MOUNT_MSDOS)
+     return("*MSDOS*");
+   else if (fs_type == MOUNT_LFS)
+     return("*LFS*");
+   else if (fs_type == MOUNT_LOFS)
+     return("*LOFS*");   
+   else if (fs_type == MOUNT_FDESC)
+     return("*FDESC*");
+   else if (fs_type == MOUNT_PORTAL)
+     return("*PORTAL*");
+   else if (fs_type == MOUNT_NULL)
+     return("*NULL*");
+   else if (fs_type == MOUNT_UMAP)
+     return("*UMAP*");
+   else if (fs_type == MOUNT_KERNFS)
+     return("*KERNFS*");
+   else if (fs_type == MOUNT_PROCFS)
+     return("*PROCFS*");
+   else if (fs_type == MOUNT_AFS)
+     return("*AFS*");
+   else if (fs_type == MOUNT_CD9660)
+     return("*CD9660*");
+   else if (fs_type == MOUNT_UNION)
+     return("*UNION*");
+   else if (fs_type == MOUNT_DEVFS)
+     return("*DEVFS*");
+   else if (fs_type == MOUNT_EXT2FS)
+     return("*EXT2FS*");
+   else if (fs_type == MOUNT_TFS)
+     return("*TFS*");
+   else
+     return("*NONE*");
+}
+#endif
+
+/* Parse File System Type Exclusion */
+int
+parse_fs_exclude(char *estr)
+{
+   char *ptr;
+   char *newestr=NULL;
+   int estr_size=0;
+   
+   /* Uppercase all characters in estr to make parsing easier.
+    * Also add '*'s around each fs type */
+   if (estr) {
+      estr_size = strlen(estr)+2;
+      newestr = malloc(estr_size+1);
+      *newestr = 0;
+      strcat(newestr,"*");
+      for (ptr = estr; *ptr != 0; ptr += 1) {
+         *ptr = toupper(*ptr);
+         if (*ptr == ',') {
+            estr_size += 1;
+            newestr = realloc(newestr,estr_size+1);
+            strcat(newestr,"*");
+         } else {
+            strncat(newestr,ptr,1);
+         }
+      }
+      strcat(newestr,"*");
+      estr = newestr;
+   }
+   
+   /* FreeBSD File System Status */
+
+#ifdef __FreeBSD__
+     {
+        struct statfs *fs_stat;
+        long bufsize=4096;
+        int num_mounts=0;
+        int i;
+        char *exclude_str=NULL;
+        
+        fs_stat = malloc(sizeof(struct statfs));
+        
+        num_mounts = getfsstat(fs_stat,bufsize,MNT_WAIT);
+        
+        for (i = 0; i < num_mounts; i+=1) {
+           if (strstr(estr,get_fs_type(fs_stat[i].f_type))) {
+              if (!exclude_str) {
+                 exclude_str = malloc(strlen(fs_stat[i].f_mntonname)+1);
+                 *exclude_str = 0;
+              } else {
+                 exclude_str = realloc(exclude_str,strlen(exclude_str)+strlen(fs_stat[i].f_mntonname)+2);
+                 strcat(exclude_str,",");
+              }
+              
+              strcat(exclude_str,fs_stat[i].f_mntonname);
+              
+           }
+        }
+        
+        if (exclude_str)
+          parse_exclude(exclude_str);
+     }
+#else
+   /* Linux File System Status */
+     {
+      char *fbuf=NULL;
+      char *head_ptr;
+      char *tail_ptr;
+      char *exclude_str=NULL;
+      char *match_str=NULL;
+   
+      fbuf = load_file(MTAB_FILE);
+      
+      if (*fbuf == 0) {
+         fbuf += 1;
+         if (!QUIET)
+           fprintf(stderr,"%s: File System Exclude: Could not open file %s: %s\n",progname,MTAB_FILE,fbuf);
+         exit(1);
+      }
+      
+      head_ptr = fbuf;
+      
+      while (head_ptr) {
+         /* find filesystem type */
+         if ((head_ptr = strchr(head_ptr,' '))) {
+            head_ptr += 1;
+            head_ptr = strchr(head_ptr,' ');
+         }
+
+         if (!head_ptr)
+           continue;
+         
+         head_ptr += 1;
+         
+         tail_ptr = strchr(head_ptr,' ');
+         if (!tail_ptr) {
+            head_ptr = NULL;
+            continue;
+         }
+         
+         *tail_ptr = 0;
+         
+         /* Check if file sytem type exists in exclude string */
+
+         match_str = realloc(match_str,strlen(head_ptr)+3);
+         match_str[0] = '*';
+         match_str[1] = 0;
+         strcat(match_str,head_ptr);
+         strcat(match_str,"*");
+         
+         for (ptr = match_str; *ptr != 0; ptr += 1)
+           *ptr = toupper(*ptr);
+         *tail_ptr = ' ';
+         
+         /* If a match is found, move back a few bytes to locate the path */
+         if (strstr(estr,match_str)) {
+            
+            head_ptr -= 2;
+            
+            while (*head_ptr != ' ' && head_ptr != fbuf)
+              head_ptr -= 1;
+                                    
+            if (head_ptr == fbuf) {
+               if (!QUIET)
+                 fprintf(stderr,"%s: File System Exclude: (1) corrupt mtab file: %s\n",progname,MTAB_FILE);
+               exit(1);                  
+            }
+            
+            head_ptr += 1;
+            
+            if (!(tail_ptr = strchr(head_ptr,' '))) {
+               if (!QUIET)
+                 fprintf(stderr,"%s: File System Exclude: (2) corrupt mtab file: %s\n",progname,MTAB_FILE);
+               exit(1);
+            }
+            
+            *tail_ptr = 0;
+            
+            /* Add the path to the exclude string */
+            if (!exclude_str) {
+               exclude_str = malloc(strlen(head_ptr)+1);
+               *exclude_str = 0;
+               strcat(exclude_str,head_ptr);
+            } else {
+               exclude_str = realloc(exclude_str,strlen(exclude_str)+strlen(head_ptr)+2);
+               strcat(exclude_str,",");
+               strcat(exclude_str,head_ptr);
+            }
+            
+            *tail_ptr = ' ';
+         }
+         
+         head_ptr = strchr(head_ptr,'\n');
+                           
+      }
+
+# ifdef DEBUG
+        printf("=-%s-=\n",exclude_str);
+# endif
+      
+      if (exclude_str)
+        parse_exclude(exclude_str);
+
+   }
+#endif   
+
+   return 1;
+}
+
+#ifndef __FreeBSD__
+
+/* Parse updatedb.conf file */
+
+void
+parse_updatedb_conf()
+{
+   char *fbuf;
+   char *head_ptr;
+   char *tail_ptr;
+   char *ptr;
+   char tmp_ch;
+   char *parse_str=NULL;
+   int times=0;
+   char var_type[11];
+   
+   fbuf = load_file(UPDATEDB_CONF);
+   
+   if (*fbuf == 0) {
+      fbuf += 1;
+/*      if (!QUIET) {
+         fprintf(stderr,"%s: could not access %s: %s\n",progname,UPDATEDB_CONF,fbuf);
+         fflush(stderr);
+      } */
+      return;
+   }
+   
+   for (times = 0; times < 2; times += 1) {
+   
+      head_ptr = fbuf;
+      
+      var_type[0] = 0;
+      
+      if (times == 0)
+        strcat(var_type,"PRUNEFS");
+      else
+        strcat(var_type,"PRUNEPATHS");        
+      
+      if ((head_ptr = strstr(fbuf,var_type))) {
+         ptr = head_ptr;
+         while (ptr != fbuf && *ptr != '\n' && *ptr != '#')
+           ptr -= 1;
+         
+         if (*ptr != '#') {
+            while (*head_ptr != '"' && *head_ptr != '\n' && *head_ptr != '\0')
+              head_ptr += 1;
+            
+            if (*head_ptr == '"') {
+               head_ptr += 1;
+               
+               /* Get each file system from PRUNEFS variable */
+               
+               while (*head_ptr != '"' && *head_ptr != 0) {
+                  
+                  while (*head_ptr != 0 && isspace(*head_ptr))
+                    head_ptr += 1;            
+                  
+                  if (*head_ptr == 0 || *head_ptr == '"')
+                    break;
+                  
+                  tail_ptr = head_ptr;
+                  while (*tail_ptr != 0 && !isspace(*tail_ptr) && *tail_ptr != '"')
+                    tail_ptr += 1;
+                  
+                  if (*tail_ptr == 0)
+                    break;
+                  
+                  tmp_ch = *tail_ptr;
+                  
+                  *tail_ptr = 0;
+                  
+                  if (!parse_str) {
+                     parse_str = malloc(strlen(head_ptr)+1);
+                     *parse_str = 0;
+                  } else {
+                     parse_str = realloc(parse_str,strlen(parse_str)+strlen(head_ptr)+2);
+                     strcat(parse_str,",");
+                  }
+                  
+                  strcat(parse_str,head_ptr);
+                  
+                  *tail_ptr = tmp_ch;
+                  head_ptr = tail_ptr;
+                  
+               }
+               
+            }
+         }
+      }
+      
+      if (parse_str) {
+         if (times == 0)     
+           parse_fs_exclude(parse_str);
+         else
+           parse_exclude(parse_str);
+      }      
+      
+      if (parse_str)
+        free(parse_str);
+      
+      parse_str = NULL;
+   }
+   
+# ifdef DEBUG
+   printf("Final Exclude: %s\n",EXCLUDE_DIR);
+   exit(0);   
+# endif
+}
+
+#endif
+
+
+/* Check to see if a path matches an excluded one. */
+
 int
 match_exclude(char *path,char *filedir) {
    int res=0;
@@ -414,147 +776,6 @@ match_exclude(char *path,char *filedir) {
    return res;
 }
 
-/* Load Dir Contents into Linked List */
-
-dir_item *
-load_dir(int dir_fd, dir_item *listptr)
-{
-   DIR *recdir;
-   struct dirent *filename = NULL;
-   struct stat statres;
-   int top=0;
-
-   if (MYfchdir(dir_fd,listptr->name,"LD_1") == -1) return(NULL);
-   
-   if ((recdir = opendir(".")) == NULL) {
-      if (!QUIET)
-          fprintf(stderr,"%s: Opendir: %s: %s\n",progname,listptr->name,strerror(errno));
-      return(NULL);
-   }
-
-   /* Open Current Directory 
-   if ((filename = readdir(recdir)) == NULL) {
-      if (!QUIET)
-          fprintf(stderr,"%s: Readdir Error: Load List: %s: %s\n",progname,listptr->name,strerror(errno));
-      closedir(recdir);
-      return(NULL);
-   }
-
-   seekdir(recdir,telldir(recdir)); */
-
-   while ((filename = readdir(recdir)) != NULL) {
-
-      while (!strcmp(filename->d_name,".") || !strcmp(filename->d_name,"..")) {
-         top=1;
-         if ((filename = readdir(recdir)) == NULL)
-             break;
-      }
-      
-      if (top) {
-
-         listptr = add_right(listptr);
-         listptr = listptr->right;
-         
-         /* If directory is emtpy then add a space to listptr */         
-         if (filename == NULL) {
-            listptr->name = malloc(1);
-            listptr->name[0] = '\0';
-            listptr->empty = 1;
-            listptr->type = 0;
-            listptr->root = 0;
-            listptr->name_len = 1;
-            closedir(recdir);
-            return(listptr);            
-         }
-         
-         top=0;
-      } else {
-         listptr = add_down(listptr);
-         listptr = listptr->down;
-      }
-      
-      /* Add information from filenames->d_name to listptr */
-      listptr->name = malloc(strlen(filename->d_name)+1);
-      listptr->name[0] = '\0';
-      strcat(listptr->name,filename->d_name);
-      listptr->name_len = strlen(listptr->name);      
-
-      if (lstat(filename->d_name,&statres) == -1) {
-         if (!QUIET)
-             fprintf(stderr,"%s: lstat: %s: %s\n",progname,filename->d_name,strerror(errno));
-      }
-      
-      if (S_ISDIR(statres.st_mode))
-          listptr->type=1;
-      else
-          listptr->type=0;
-      
-      listptr->root = 0;
-
-   }
-   
-   closedir(recdir);
-   
-   return(listptr);
-}
-
-/* Get Current Directory */
-char *
-get_cur_dir(dir_item *listptr, char *dirstr)
-{
-   
-   dir_item *dir_ptr;
-   char **dirarray;
-   int items=0;
-   int i;
-   int grow;
-   
-
-   if (cur_dir != NULL) {
-      free(cur_dir);
-   }
-
-   dir_ptr = listptr;
-   
-   while (!dir_ptr->root) {
-      dir_ptr=dir_ptr->left;
-      items++;
-   }
-
-   dirarray = calloc(items,sizeof(char *));
-   
-   dir_ptr=listptr->left;
-   
-   for (i = items-1; i != -1; i--) {
-      dirarray[i] = malloc(strlen(dir_ptr->name)+1);
-      dirarray[i][0] = '\0';
-      strcat(dirarray[i],dir_ptr->name);
-
-      dir_ptr=dir_ptr->left;
-   }
-
-   cur_dir = malloc(1);
-   cur_dir[0] = '\0';
-   grow=1;
-   
-   for (i = 0; i != items; i++) {
-      grow+=strlen(dirarray[i])+1;
-      cur_dir = realloc(cur_dir,grow);
-      strcat(cur_dir,dirarray[i]);
-      if ((strcmp(dirstr,"/") != 0 && i == 0) || i > 0)
-          strcat(cur_dir,"/");
-   }
-
-   for (i = 0; i != items; i++) {
-      free(dirarray[i]);
-   }
-   
-   free(dirarray);
-
-   
-   return(cur_dir);
-}
-
 /* FRCODE - Incremental Encoding algorithm */
 
 int
@@ -636,20 +857,19 @@ create_db(char *dirstr)
 {
    int res=0;
    struct stat statres;
-   int dir_fd;
-   dir_item *listptr=NULL;
-   dir_item *tmpptr=NULL;
-   int load;
-   int jump=0;
-   int list_down=0;
    FILE *fd;
-
-   if (UID != SLOC_UID) {
-      if (!QUIET)
-          fprintf(stderr,"%s: You are not authorized to create an slocate database!\n",progname);
-      return(0);
-   }
+   FTS *dir;
+   FTSENT *file;
+   int i;
+   char **dirchk;
    
+
+   if (!NEWOUTPUT && UID != SLOC_UID) {
+      if (!QUIET)
+          fprintf(stderr,"%s: You are not authorized to create a default slocate database!\n",progname);
+      return(1);
+   } 
+
    fd = fopen(TMPSLOCATEDB,"w");
    putc(slevel,fd);
    
@@ -657,151 +877,49 @@ create_db(char *dirstr)
       /* Make sure there are no problems accessing the source directory */
       if (lstat(dirstr,&statres) == -1) {
          if (!QUIET)
-             fprintf(stderr,"%s: lstat: d%s/: %s\n",progname,dirstr,strerror(errno));
+      fprintf(stderr,"%s: lstat: d%s/: %s\n",progname,dirstr,strerror(errno));
     fclose(fd);
-         return(-1);
+         return(1);
+      }
+      
+      if (strlen(dirstr) > 1) {
+    if (dirstr[strlen(dirstr)-1] == '/')
+      dirstr[strlen(dirstr)-1] = 0;        
       }
    }
-   
-   /* Initialize 2D Linked List */
-   listptr = init_2D_list();
-
-   if (dirstr) {
-       if (dirstr[strlen(dirstr)-1] == '/') dirstr[strlen(dirstr)-1] = '\0';
-   } else {
+   else {
       dirstr = malloc(2);
-      dirstr[0]='/';
-      dirstr[1]='\0';
-   }
-
-   
-   listptr->name = malloc(strlen(dirstr)+1);
-   listptr->name[0] = '\0';
-   strcat(listptr->name,dirstr);
-   listptr->st_uid = statres.st_uid;
-   listptr->st_gid = statres.st_gid;
-   listptr->st_mode = statres.st_mode;
-   listptr->root = 1;
-   listptr->empty=0;
-   
-   
-   /* Open Source Directory */
-   if ((dir_fd = open(dirstr,O_RDONLY)) == -1) {
-      if (!QUIET)
-          fprintf(stderr,"%s: open: %s: %s\n",progname,dirstr,strerror(errno));
-      fclose(fd);
-      return(-1);
+      *dirstr = 0;
+      strcat(dirstr,"/");
    }
    
-   /* Change CWD to source dir */
-   if (!MYfchdir(dir_fd,dirstr,"CDB_1")) {
-      close(dir_fd);
-      fclose(fd);
-      return(0);
-   }
+   dirchk = malloc(sizeof(char **)*2);
    
-   if (EXCLUDE && match_exclude(dirstr,""))
-       load = 0;
-   else
-       load = 1;
-
-   while (!jump) {
-      if (load) {
-         /* Load directory contents into linked list */
-         listptr = load_dir(dir_fd,listptr);
-         load=0;
-         listptr=listptr->left->right;
+   *dirchk = dirstr;
+   dirchk[1] = NULL;
+   
+   dir = fts_open(dirchk,FTS_PHYSICAL,NULL);   
+   
+   /* The new FTS() funtionality */
+   
+   for (i = 0; i > -1; i+=1) {
+      file = fts_read(dir);
+      
+      if (!file)
+   break;
+      
+      if (file->fts_info != FTS_DP && file->fts_info != FTS_NS) {
+    
+    if ((EXCLUDE && !match_exclude(file->fts_path,"")) || !EXCLUDE)
+      frcode(fd,file->fts_path,"");
+    else {
+       fts_set(dir,file,FTS_SKIP);
+    }
       }
       
-      if (!listptr->empty) {
-
-         tmpptr = listptr;
-         if (list_down) {
-            listptr=listptr->down;
-         }
-         list_down=0;         
-         while (listptr != NULL) {
-            if (!EXCLUDE || (EXCLUDE && !match_exclude(get_cur_dir(listptr,dirstr),listptr->name))) {               
-               frcode(fd,get_cur_dir(listptr,dirstr),listptr->name);               
-            }
-            if (!listptr->type) {
-               listptr=listptr->down;
-            } else {
-               load=1;
-               break;
-            }
-         }
-      } else {
-         if (!MYchdir("..","create_db")) {
-            close(dir_fd);
-       fclose(fd);
-            return(1);
-         }         
-         listptr=listptr->left;
-         free_right(listptr);
-         tmpptr=listptr;
-         listptr=listptr->down;
-         if (listptr != NULL)
-             load = listptr->type;
-         list_down=0;
-      }
-      
-      if (listptr == NULL) {
-         
-         while (listptr == NULL) {
-            listptr=tmpptr;
-            if (listptr->root) {
-               jump=1;
-               break;
-            }
-            listptr=listptr->left;
-            
-            if (!MYchdir("..","CDB_2")) {
-               close(dir_fd);
-          fclose(fd);
-               return(1);
-            }
-            
-            close(dir_fd);
-            
-            if ((dir_fd = open(".",O_RDONLY)) == -1) {
-               if (!QUIET)
-                   fprintf(stderr,"%s: open: CDB_3: %s\n",progname,strerror(errno));
-          fclose(fd);
-               return(1);
-            }
-            
-            tmpptr=listptr;
-            free_right(listptr);
-            listptr=listptr->down;            
-         }
-         
-         load=0;
-
-      } else if (listptr->type) {
-         
-         /* Change to new directory */
-         if ((EXCLUDE && match_exclude(get_cur_dir(listptr,dirstr),listptr->name) == 1) || (chdir(listptr->name) == -1)) {
-            load=0;
-            list_down=1;
-         } else {
-            
-            close(dir_fd);
-            
-            if ((dir_fd = open(".",O_RDONLY)) == -1) {
-               if (!QUIET)
-                 fprintf(stderr,"%s: open: CDB_5: %s: %s%s\n",progname,strerror(errno),
-                         get_cur_dir(listptr,dirstr),listptr->name);
-               //   return(1);  Don't bail just because we can't read a directory.
-               // This might happen even as root on some NFS partitions.
-               load=0;
-               list_down=0;          
-            }
-         }
-                  
-      }
-
-   }   
+   }
+   
+   fts_close(dir);
    
    fclose(fd);
    chdir(prog_CWD);
@@ -839,6 +957,7 @@ decode_db(char *database, char *str)
    int nmatch=32;
    regmatch_t pmatch[32];
    int cur_queries = 0;
+   int reg_res;
    
    if ((fd = fopen(database,"r")) == NULL) {
       if (!QUIET)
@@ -917,8 +1036,8 @@ decode_db(char *database, char *str)
       if (REGEXP) {
          globflag=0;
          preg = malloc(sizeof(regex_t));
-         if ((res = regcomp(preg,regexp,0)) != 0) {
-            regerror(res, preg, errbuf,1024);
+         if ((reg_res = regcomp(preg,regexp,0)) != 0) {
+            regerror(reg_res, preg, errbuf,1024);
             if (!QUIET)
               fprintf(stdout,"error: %s: regular expression: %s\n",progname,errbuf);
             exit(1);
@@ -949,6 +1068,7 @@ decode_db(char *database, char *str)
         regfree(preg);
       
       if (printit) {
+         res = 0;
          cur_queries += 1;
          printf("%s\n",codedpath);
          if (max_queries > 0 && cur_queries == max_queries)
@@ -958,7 +1078,7 @@ decode_db(char *database, char *str)
    
    fclose(fd);
    
-   return(1);
+   return(res);
 }
 
 /* Main Function */
@@ -980,8 +1100,12 @@ main(int args, char **argv)
 
    progname = ((p = strrchr(argv[0],'/')) ? p+1 : *argv);     
    
-   if (!strcmp(progname,"updatedb"))
-     ROOTDIR=1;
+   if (!strcmp(progname,"updatedb")) {
+      ROOTDIR=1;
+      #ifndef __FreeBSD__
+        parse_updatedb_conf();
+      #endif
+   }
    
    /* Get Current Working Directory */
    
@@ -1002,27 +1126,34 @@ main(int args, char **argv)
    UID = getuid();        
    GID = getgid();
 
-   parse_create_path(SLOCATEDB);
    parse_decode_path(SLOCATEDB);
    parse_decode_path(getenv("LOCATE_PATH"));
 
-   while ((ch = getopt(args,argv,"VvuhqU:r:o:e:l:d:-:n:")) != EOF) {
+   while ((ch = getopt(args,argv,"VvuhqU:r:o:e:l:d:-:n:f:c")) != EOF) {
       switch(ch) {
-      case 'h':
+       case 'h':
          usage();
          break;
-      case 'q':
+       case 'q':
          QUIET=1;
          break;
-      case 'V':
+       case 'V':
          printf("%s",VERSION);
-         exit(1);
+         exit(0);
          break;
-      case 'v':
+       case 'v':
          VERBOSE=1;
          break;
-      case 'e':
-         EXCLUDE_DIR = parse_exclude(optarg);
+       case 'e':
+         parse_exclude(optarg);
+         break;
+       case 'f':
+         parse_fs_exclude(optarg);
+         break;
+       case 'c':
+#ifndef __FreeBSD__
+         parse_updatedb_conf();
+#endif
          break;
        case 'l':
          slevel = optarg[0];
@@ -1032,10 +1163,10 @@ main(int args, char **argv)
             exit(1);
          }
          break;
-      case 'u':
+       case 'u':
          ROOTDIR=1;
          break;
-      case 'U':
+       case 'U':
          ROOTDIR=0;
          SPECDIR=1;
          spec_dir = malloc(strlen(optarg)+1);
@@ -1048,15 +1179,16 @@ main(int args, char **argv)
          *regexp = 0;
          strcat(regexp,optarg);
          while ((database = SLOCATE_PATH[i++])) res |= decode_db(database, optarg);
-         return(!res);
+         return(res);
          break;
-      case 'o':
+       case 'o':
          parse_create_path(optarg);
+         NEWOUTPUT=1;
          break;
-      case 'd':
+       case 'd':
          parse_decode_path(optarg);
          break;
-      case '-':
+       case '-':
          parse_dash(optarg);
          break;
        case 'n':
@@ -1068,7 +1200,7 @@ main(int args, char **argv)
          }
          max_queries = atoi(optarg);
          break;
-      default:
+       default:
          return(1);
          break;
          
@@ -1092,5 +1224,5 @@ main(int args, char **argv)
    } else
        usage();
 
-   return(!res);
+   return(res);
 }
